import unittest
from pygenesig.bioqc import *
from pygenesig.validation import *
from pygenesig.tools import load_gmt
from rpy2.robjects.packages import importr
import random
from . import TESTDATA

base = importr("base")


class TestBioQC(unittest.TestCase):
    def setUp(self):
        # np.matrix does NOT work l
        self.expr = np.array(
            np.matrix("5. 5. 0. 0.; " "0. 0. 5. 5.;" "3. 2. 2. 2.;" "0. 0. 0. 1.")
        )
        self.target = ["A", "A", "B", "B"]
        self.signatures = {  # signatures used in csv and gmt test files.
            "sig_a": [1, 2, 3, 4],
            "sig_b": [5, 6, 7, 8],
            "sig_c": [9, 10, 11, 12],
        }

    def test_signatures2gmt(self):
        gmt = BioQCSignatureTester.signatures2gmt(self.signatures)
        self.assertCountEqual(self.signatures.keys(), list(base.names(gmt)))

    def test_log_pvalue(self):
        """check if log pvalues are identical with running BioQC in R.
        Testcase generated with test_bioqc_log_pvalue.R
        """
        log_p_expected = np.loadtxt(
            TESTDATA / "./bioqc/test_bioqc_log_pvalue.csv",
            delimiter=",",
            skiprows=1,
            usecols=range(1, 5),
        )
        gmt = bioqc.readGmt(str(TESTDATA / "./bioqc/test_bioqc_log_pvalue.gmt"))
        gene_symbols = [str(x) for x in range(self.expr.shape[0])]
        p_actual = BioQCSignatureTester.run_bioqc(self.expr, gene_symbols, gmt)
        log_p_actual = -np.log10(p_actual)
        np.testing.assert_array_almost_equal(log_p_expected, log_p_actual)

    def test_score_signatures(self):
        log_p_expected = np.loadtxt(
            TESTDATA / "./bioqc/test_bioqc_log_pvalue.csv",
            delimiter=",",
            skiprows=1,
            usecols=range(1, 5),
        )
        st = BioQCSignatureTester(self.expr, self.target)
        log_p = st.score_signatures(self.signatures)
        # works, because the GMT is properly sorted.
        np.testing.assert_array_almost_equal(log_p_expected, log_p)

    def test_test_with_good_signatures(self):
        signatures_good = {"A": [0], "B": [1]}
        tester = BioQCSignatureTester(self.expr, self.target)
        actual, predicted = tester._test_signatures(signatures_good, np.array(range(4)))
        np.testing.assert_array_equal(self.target, predicted)

    def test_test_with_bad_signatures_and_subset(self):
        signatures_bad = {"A": [3, 2], "B": [2]}
        tester = BioQCSignatureTester(self.expr, self.target)
        actual, predicted = tester._test_signatures(signatures_bad, np.array([0, 1, 3]))
        cm = tester.confusion_matrix(signatures_bad, actual, predicted)
        cm_expected = np.matrix("0 2;" "0 1")
        np.testing.assert_array_equal(cm_expected, cm)

    def test_order_of_signatures(self):
        signatures_all = {"A": [0], "B": [1], "C": [3], "D": [3], "E": [3], "F": [3]}
        tester = BioQCSignatureTester(self.expr, self.target)
        cm_expected = np.matrix(
            "2 0 0 0 0 0;"
            "0 2 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0"
        )
        for i in range(10):
            """test in loop to have random permutations of the dictionary."""
            new_dict = {k: v for k, v in signatures_all.items()}
            actual, predicted = tester._test_signatures(new_dict, np.array(range(4)))
            cm = tester.confusion_matrix(signatures_all, actual, predicted)
            np.testing.assert_array_equal(cm_expected, cm)

    def test_empty_signature_all(self):
        signatures_bad = {"A": [], "C": [], "B": []}
        tester = BioQCSignatureTester(self.expr, self.target)
        with self.assertRaises(SignatureTesterException):
            actual, predicted = tester._test_signatures(
                signatures_bad, np.array([0, 1, 3])
            )

    def test_empty_signature1(self):
        signatures_bad = {"A": [3], "C": [], "B": []}
        tester = BioQCSignatureTester(self.expr, self.target)
        actual, predicted = tester._test_signatures(signatures_bad, np.array([0, 1, 3]))
        cm = tester.confusion_matrix(signatures_bad, actual, predicted)
        cm_expected = np.matrix("2 0 0;" "1 0 0;" "0 0 0")
        np.testing.assert_array_equal(cm_expected, cm)

    def test_empty_signatures_order(self):
        signatures_all = {"A": [], "B": [2], "C": [3], "D": [], "E": [], "F": [3]}
        tester = BioQCSignatureTester(self.expr, self.target)
        cm_expected = np.matrix(
            "0 2 0 0 0 0;"
            "0 2 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0"
        )
        for i in range(10):
            """test in loop to have random permutations of the dictionary."""
            new_dict = {k: v for k, v in signatures_all.items()}
            actual, predicted = tester._test_signatures(new_dict, np.array(range(4)))
            cm = tester.confusion_matrix(signatures_all, actual, predicted)
            np.testing.assert_array_equal(cm_expected, cm)

    def test_dtype(self):
        expr = np.array([["1.5", "2.7"], ["2.8", "3.7"]])
        target = np.array(["A", "B"])
        with self.assertRaises(TypeError):
            st = BioQCSignatureTester(expr, target)

    def test_reorder(self):
        template = ["A", "B", "C", "D"]
        target = ["D", "B", "C", "A"]
        np.testing.assert_array_equal(
            [3, 1, 2, 0], BioQCSignatureTester._reorder(template, target)
        )

    def test_reorder_with_random_data(self):
        template = np.array(list(range(1000)))
        target = list(range(1000))
        random.shuffle(target)
        target = np.array(target)
        reorder = BioQCSignatureTester._reorder(template, target)
        np.testing.assert_array_equal(template, target[reorder])



"""
The Gini-coefficient (gini index) measures the inequality among values of
a frequency distribution. Originally applied to measure income inequality
we can also use it on gene expression data to find genes that are
over-represented in certain samples.
"""

import numpy as np
import pandas as pd
from pygenesig.validation import SignatureGenerator
from pygenesig.tools import collapse_matrix
from pygenesig.gini import gini


def get_gini_signatures(df_aggr, min_gini=0.7, max_rk=3, min_expr=1):
    """
    Generate gene signatures using gini index.

    Finds over-represented genes for each sample group, given the specified thresholds.

    Args:
        df_aggr (pd.DataFrame): m x n pandas DataFrame with m Genes and n tissues
        min_gini (float): gini cutoff, genes need to have a gini index larger than this value.
        max_rk (int): rank cutoff, include genes if they rank <= max_rank among all tissues.
        min_expr (float): minimal expression

    Returns:
        dict of list: A signature dictionary.

        Example::

            {
                "tissue1" : [list, of, gene, ids],
                "tissue2" : [list, of, other, genes],
                ...
            }

    """
    signatures = {tissue: [] for tissue in df_aggr}
    for i in range(df_aggr.shape[0]):
        row = df_aggr.iloc[i, :].sort(inplace=False, ascending=False)
        if row.iloc[0] >= min_expr:
            while gini(row) >= min_gini:
                signatures[row.index[0]].append(i)
                row.drop(row.index[0], inplace=True)
    signatures = {tissue: set(genes) for tissue, genes in signatures.items()}
    return signatures


class CleverGiniSignatureGenerator(SignatureGenerator):
    """
    Use gini index to generate gene signatures.

    Genes, which are specific for a tissue result in a high gini index,
    whereas genes equally present in all tissues have a gini index close to zero.

    The idea is, that genes with a high gini index will reliably identify
    their tissue of origin.

    Args:
        expr (np.ndarray): m x n matrix with m samples and n genes
        target (array-like): m-vector with true tissue for each sample
        min_gini (float): gini cutoff, genes need to have a gini index larger than this value.
        max_rk (int): rank cutoff, include genes if they rank ``<= max_rank`` among all tissues.
        min_expr (float): genes need to have at least an expression ``>= min_expr`` to be included.
        aggregate_fun (function): function used to aggregate samples of the same tissue.
    """

    def __init__(
        self, expr, target, min_gini=0.7, max_rk=3, min_expr=1, aggregate_fun=np.median
    ):
        super(CleverGiniSignatureGenerator, self).__init__(expr, target)
        self.min_gini = min_gini
        self.max_rk = max_rk
        self.min_expr = min_expr
        self.aggregate_fun = aggregate_fun

    def _mk_signatures(self, expr, target):
        df_aggr = collapse_matrix(
            self.expr, self.target, axis=1, aggregate_fun=self.aggregate_fun
        )
        return get_gini_signatures(
            df_aggr, min_gini=self.min_gini, max_rk=self.max_rk, min_expr=self.min_expr
        )



"""
pygenesig's main module: Contains the abstract base classes for signature
generation and testing and provides tools for cross-validation.
"""


from abc import abstractmethod, ABCMeta
import numpy as np
import sklearn.model_selection
from dask import delayed
import os.path
from collections import Counter


def filter_samples(exprs, target, n_splits):
    """
    Remove all categories and the respective samples
    from the input data that have not at least *n_splits*
    samples per category.

    This is useful when performing a cross-validation with
    *n* folds, as we need at least *n* samples per category.

    Args:
        exprs:
        target:
        n_splits:

    Returns:
        exprs:
        target:

    """
    target_cnt = Counter(target)
    keep_tissues = [tissue for tissue, count in target_cnt.items() if count >= n_splits]
    target_mask = np.in1d(target, keep_tissues)
    return exprs[:, target_mask], target[target_mask]


def cv_score(
    expr_file,
    target_file,
    signature_generator,
    signature_tester,
    splitter=sklearn.model_selection.StratifiedKFold(n_splits=10),
    sg_kwargs={},
    st_kwargs={},
):
    """
    Perform a crossvalidation by generating and testing signatures on a given expression matrix.

    Args:
        expr_file (str): path to numpy.array (binary) file containing expression matrix saved with ``numpy.save()``
        target_file (str): path to plaintext file containing target classes, line-by-line
        signature_generator (SignatureGenerator): ``SignatureGenerator`` used to derive the signatures
        signature_tester (SignatureTester): ``SignatureTester`` used to check the quality of signatures
        splitter (sklearn.model_selection._BaseKFold): crossvalidation method from `scikit-learn`_.

    Returns:
        (list, list, list, list): list of signature dictionaries containing the signatures generated
        on each fold, list of confusion matrices containing the confusion matrices calculated on each fold,
        list of indices used for training for each fold, list of indices used for testing for each fold.

    .. _scikit-learn:
        http://scikit-learn.org/stable/modules/classes.html#module-sklearn.model_selection

    """
    # we need the number of samples locally for splitting in test and training sets
    target_file = os.path.realpath(target_file)
    expr_file = os.path.realpath(expr_file)
    target_local = np.genfromtxt(target_file, dtype=str, delimiter=",")

    # the rest we can run delayed on the dask cluster
    signatures = []
    scores = []
    train_list = []
    test_list = []
    for train, test in splitter.split(list(enumerate(target_local)), target_local):
        # due to some bug in dask, it's faster to load all files on every worker separately.
        train_list.append(train)
        test_list.append(test)
        expr = delayed(np.load)(expr_file)
        target = delayed(np.genfromtxt)(target_file, dtype=str, delimiter=",")
        sg = delayed(signature_generator)(expr, target, **sg_kwargs)
        st = delayed(signature_tester)(expr, target, **st_kwargs)
        signature = delayed(sg.mk_signatures)(train)
        signatures.append(signature)
        score = delayed(st.score_signatures, nout=2)(signature, test)
        scores.append(score)
    sig_list = delayed(list)(signatures)
    res_list = delayed(list)(scores)
    return sig_list, res_list, train_list, test_list


class SignatureGenerator(metaclass=ABCMeta):
    """
    Abstract base-class for generating gene signatures.
    Child classes build gene signatures from a gene expression matrix and target labels.

    When initializing ``SignatureGenerator`` the input data is checked for consistency.
    You can override the ``__init__`` method in your implementation of ``SignatureGenerator``
    to pass additional parameters, however, make always sure to call this method
    in your ``__init__`` function to take advantage of the consistency check::

        def __init__(self, expr, target, your_param):
            super(YourSignatureGenerator, self).__init__(expr, target)
            # your code goes here


    Args:
        expr (np.ndarray): m x n matrix with m samples and n genes
        target (array-like): m-vector with true tissue for each sample

    .. automethod:: _mk_signatures
    """

    def __init__(self, expr, target):
        if not expr.shape[1] == len(target):
            raise ValueError(
                "The length of target must equal the number of samples (columns) in the expr matrix. "
            )
        if not np.issubdtype(expr.dtype, int) and not np.issubdtype(expr.dtype, float):
            raise TypeError("Expression needs to be numeric. (dtype=int | dtype=float)")

        self.expr = expr
        self.target = target

    @abstractmethod
    def _mk_signatures(self, expr, target):
        """
        Implement this method.
        Generates signatures for a list of tissues given a gene expression matrix.
        This method is called internally by the public method `mk_signatures`, which
        takes care of subsetting the expression data.

        Args:
            expr: m x n matrix with m samples and n genes
            target: vector with true tissue for each sample

        Returns:
            dict: tissue_name -> [list, of, enriched, genes]. The indices correspond to expr.index.

        Note:
            When implementing this method, make sure, that every tissue in `this.target` has an entry in the
            dictionary, even when the signature does not contain any genes.

        """
        pass

    def mk_signatures(self, subset=None):
        """
        Make gene signatures on a subset of the samples in the gene expression matrix.

        Args:
            subset: Sample indices (columns of expression matrix) to use for signature generation.
                Useful for cross-validation. If omitted, all samples will be used.

        Returns:
            dict: signature dictionary. Maps target labels to a list of row-indices of the gene
            expression matrix.

            Example::

                signatures = {
                    'adipose tissue' : [
                        42,
                        124,
                        256,
                        1038,
                        ... ],
                    'skeletal muscle' : [
                        52,
                        679,
                        ... ],
                    ...
                }


        """
        if subset is None:
            subset = np.array(list(range(len(self.target))))

        return self._mk_signatures(self.expr[:, subset], self.target[subset])


class SignatureTester(metaclass=ABCMeta):
    """
    Abstract base-class for testing gene signatures.
    Child classes test if signatures are able to identify their respective samples,
    given an expression matrix and a list of the actual tissues.

    When initializing ``SignatureTester`` the input data is checked for consistency.
    You can override the ``__init__`` method in your implementation of ``SignatureTester``
    to pass additional parameters, however, make always sure to call this method
    in your ``__init__`` function to take advantage of the consistency check::

        def __init__(self, expr, target, your_param):
            super(YourSignatureTester, self).__init__(expr, target)
            # your code goes here

    Args:
        expr (np.ndarray): m x n gene expression matrix with m samples and n genes
        target (array-like): m-vector with true tissue for each sample (`ground truth`)

    .. automethod:: _score_signatures
    """

    def __init__(self, expr, target):
        if not expr.shape[1] == len(target):
            raise ValueError(
                "The length of target must equal the number of samples (columns) in the expr matrix. "
            )
        if not np.issubdtype(expr.dtype, int) and not np.issubdtype(expr.dtype, float):
            raise TypeError("Expression needs to be numeric. (dtype=int | dtype=float)")
        self.expr = expr
        self.target = np.array(target)

    @staticmethod
    def sort_signatures(signatures):
        """
        Retrieve the signatures in a consistent order.

        The signature dictionary is a python default dict, which is unordered.
        However, when displaying results, it is desirable to display the
        signatures in a consistent order.

        You can use this method to iterate over the signature dictionary::

            for tissue in sort_signatures(signatures):
                print(tissue, signatures[tissue])

        Confusion matrices generated with ``SignatureTester.confusion_matrix`` are
        also sorted with ``sort_signatures``.

        Args:
            signatures: signature dictionary.

        Returns:
            list: sorted keys of the signature dictionary.

        """
        return sorted(signatures.keys())

    @abstractmethod
    def _score_signatures(self, expr, signatures):
        """
        Implement this method.

        Generates a score for each sample and signature.

        Args:
            expr (np.ndarray): m x n gene expression matrix with m genes and n samples
            signatures (dict of list): Signatures dictionary with j signatures.
                Dictionary: tissue_name -> [list, of, enriched, genes].

        Returns:
            np.ndarray: j x n score matrix with j signatures and n samples containing the
            scores generated by this method for each sample and signature. The rows are sorted by
            ``SignatureGenerator.sort_signatures``.

        .. important::
            Make sure the rows in the result matrix are sorted by ``SignatureGenerator.sort_signatures``.
            Make sure that a row for each signature in ``signatures`` is included in the matrix, even
            *if the signatures is empty* (number of genes = 0).

        """
        pass

    def score_signatures(self, signatures, subset=None):
        """
        Generates a score for each sample and signature. A high score represents
        a high representation of the signature in the sample.

        Args:
            signatures (dict of list): Signatures dictionary with j signatures.
                Dictionary: tissue_name -> [list, of, enriched, genes].
            subset: sample indices (columns of expression matrix) to use for testing. Useful for crossvalidation.

        Returns:
            np.ndarray: j x n score matrix with j signatures and n samples containing the
            scores generated by this method for each sample and signature. The rows are sorted by
            ``SignatureGenerator.sort_signatures``.

        """
        if all([len(genes) == 0 for genes in signatures.values()]):
            raise SignatureTesterException(
                "Signature Set contains only empty signatures. "
            )

        if subset is None:
            # take all
            subset = np.array(list(range(len(self.target))))

        return self._score_signatures(self.expr[:, subset], signatures)

    def classify(self, signatures, score_matrix, subset=None):
        """
        Test signatures based on the expression matrix. Predicts the class labels
        using the signatures. A sample is predicted as the label associated with
        the highest scoring signature.

        Args:
            signatures (dict[str, list]): Signatures dictionary returned by ``SignatureGenerator.mk_signature``.
            score_matrix: j x n score matrix with j signatures and n samples generated by ``SignatureTester.score_signatures``
            subset: sample indices (columns of expression matrix) to use for testing. Useful for crossvalidation.

        Returns:
            (list, list): list of actual labels and predicted labels.

        """
        if subset is None:
            # take all
            subset = np.array(list(range(len(self.target))))

        actual = self.target[subset]  # standard of truth
        signature_list = self.sort_signatures(signatures)
        predicted = [signature_list[m] for m in np.nanargmax(score_matrix, axis=0)]

        return actual, predicted

    def confusion_matrix(self, signatures, actual, predicted):
        """
        Make a confusion matrix.

        This is a wrapper for the ``sklearn.metrics.confusion_matrix``. Makes the matrix contain all
        labels available in the signature.

        Note:
            The rows and columns are sorted with ``SignatureTester.sort_signatures``.
            You can therefore use ``sort_signatures`` to label you confusion matrix.

        Args:
            signatures: dict of signatures that was used to predict the labels
            actual: list of actual labels
            predicted: list of predicted labels

        Returns:
            numpy.matrix: confusion matrix

        """
        return sklearn.metrics.confusion_matrix(
            actual, predicted, labels=self.sort_signatures(signatures)
        )

    def _test_signatures(self, signatures, subset):
        """
        Wrapper for score_signatures and classify. Used for unit testing.
        """
        scores = self.score_signatures(signatures, subset)
        return self.classify(signatures, scores, subset)


class SignatureTesterException(Exception):
    pass



import pandas as pd
import numpy as np
import itertools


def _read_flatfile(input_file):
    """read array from file, one entry per line"""
    with open(input_file) as f:
        output_array = np.array(f.read().splitlines())
    return output_array


def _write_flatfile(output_file, input_array):
    """write array to file, one entry per line"""
    with open(output_file, "w") as f:
        for e in input_array:
            f.write(e + "\n")


#################################################################
# expression
#################################################################
def write_expr(expression_matrix, file):
    """Store a m x n gene expression matrix as numpy object."""
    np.save(file, expression_matrix)


def read_expr(expr_file):
    """Read a m x n gene expression matrix from a numpy object."""
    return np.load(expr_file)


def read_gct(file):
    """
    Read a `GCT file`_ to a gene expression matrix.

    Args:
        file (str): path to GCT file

    Returns:
        np.array: gene expression matrix.

    .. _GCT file:
        http://software.broadinstitute.org/cancer/software/genepattern/file-formats-guide#gct

    """
    gct = pd.read_csv(file, sep="\t", skiprows=2, index_col=0)
    exprs = gct.iloc[:, 1:].values  # get rid of description column
    return exprs


def write_gct(file, exprs, samples=None, description=None, name=None):
    """
    Write a gct file.

    Args:
        file (str): path to output file
        exprs (np.array): m x n matrix with m genes and n samples
        samples: n-array with the labels for the n samples
        description: m array with a description for each gene (e.g. gene symbol)
        name: m array with the name for each gene (e.g. gene index)

    """
    if description is None:
        description = np.repeat("na", exprs.shape[0])
    if name is None:
        name = np.arange(0, exprs.shape[0])
    if samples is None:
        samples = np.arange(0, exprs.shape[1])
    assert exprs.shape[0] == description.size == name.size
    assert exprs.shape[1] == samples.size

    gct = pd.DataFrame(exprs)
    gct.columns = samples
    fdata = pd.DataFrame({"NAME": name, "Description": description})
    gct = pd.concat((fdata, gct), axis=1)
    gct.set_index("NAME", inplace=True)

    with open(file, "w") as f:
        f.write("#1.2\n")
        f.write("{} {}\n".format(*exprs.shape))

    gct.to_csv(file, mode="a", sep="\t")


##################################################################
# target (=tissue annotations)
##################################################################
def write_target(target_array, file):
    """Given a m x n gene expression matrix with m genes and n samples. Write an n-array with
    one target annotation for each sample."""
    _write_flatfile(file, target_array)


def read_target(target_file):
    """Given a m x n gene expression matrix with m genes and n samples. Read an n-array with
    one target annotation for each sample."""
    return _read_flatfile(target_file)


##################################################################
# feature (=gene symbol annotation)
##################################################################
def write_rosetta(rosetta_array, rosetta_file):
    """Alias for `write_feature`.

    Given a m x n gene expression matrix with m genes and n samples. Write a m-array
    with one identifier for each gene.

    This can be used to map the index-based signature back to gene symbols."""
    _write_flatfile(rosetta_file, rosetta_array)


def read_rosetta(rosetta_file, as_dict=True, inverse=False):
    """Given a m x n gene expression matrix with m genes and n samples. Read an m-array
    with one identifier for each gene.

    If `as_dict` is True, this will be converted into a dictionary mapping the gene-index
    to the gene identifier. This can be used to map the index-based signatures
    back to gene symbols.

    If `inverse` is True, the mapping is inverse, and the gene-symbol will be mapped to the
    gene-index.

    Args:
        rosetta_file (str): path to file
        as_dict (boolean): If True, a dictionary will be returned. Else it will be a flat numpy.array.
        inverse (boolean): If true, map gene-symbol to index.

    .. Important::
        Be carefule when using `inverse = True` when the list in
        rosetta_file is not unique. In that case only the last
        entry makes it into the list!

    Returns:
        dict or np.array: mapping or numpy array.

    """
    fdata = _read_flatfile(rosetta_file)
    if as_dict:
        return make_rosetta_dict(fdata, inverse)
    else:
        return fdata


def make_rosetta_dict(array, inverse=False):
    """
    convert an array to a dictonary mapping the index to the corresponding array entry.
    Use `inverse` to reverse the mapping, i.e. the array entry to the index.

    Args:
        array (array-like):
        inverse (boolean):

    Returns:
        dict: the map

    """
    if inverse:
        return {
            # '-' is an artifact from ribiosAnnotation
            gene_symbol: i
            for i, gene_symbol in enumerate(array)
            if gene_symbol != "-"
        }
    else:
        return dict(enumerate(array))


##################################################################
# signatures
##################################################################
def write_gmt(signatures, file, description="na"):
    """
    Writes signatures to a `GMT file`_.

    Args:
        signatures (dict of list): signature dictionary
        file: path to output file
        description: text to fill in the gmt description field.

    .. _GMT file:
        http://software.broadinstitute.org/cancer/software/gsea/wiki/index.php/Data_formats#GMT:_Gene_Matrix_Transposed_file_format_.28.2A.gmt.29

    """
    with open(file, "w") as f:
        for sig, genes in sorted(signatures.items()):
            genes = sorted([str(g) for g in signatures[sig]])
            f.write("\t".join(itertools.chain([sig, description], genes)) + "\n")


def load_gmt(file):
    """
    Deprecated.
    Alias for read_gmt.
    """
    return read_gmt(file)


def read_gmt(file):
    """
    Read a `GMT file`_ into a signature dictionary.

    Args:
        file: path to GMT file

    Returns:
        dict of list: signature dictionary

        Example::

            {
                "tissue1" : [list, of, gene, ids],
                "tissue2" : [list, of, other, genes],
                ...
            }

    .. _GMT file:
        http://software.broadinstitute.org/cancer/software/gsea/wiki/index.php/Data_formats#GMT:_Gene_Matrix_Transposed_file_format_.28.2A.gmt.29


    """
    signatures = {}
    with open(file) as f:
        for line in f.readlines():
            cols = line.strip().split("\t")
            name = cols[0]
            genes = cols[2:]
            signatures[name] = genes
    return signatures



"""
`MCPCounter`_ is a method described by Becht et al., Genome Biology (2015)
for deconvolution of tissue-infiltrating immune and stromal cell populations.

Besides presenting a method for gene-expression based deconvolution they put a
lot of effort into curating signatures.

This module

* re-implements MCP counter in python as SignatureTester
* implements a SignatureGenerator based on the MCP method for curating signatures.

.. _MCPCounter:
    http://dx.doi.org/10.1186/s13059-016-1070-5
"""

from pygenesig.validation import SignatureGenerator, SignatureTester
import numpy as np
from sklearn.metrics import roc_auc_score


def fold_change(expr, positive_mask):
    """
    Compute the fold change between the
    positive and negative samples for a single gene `G`.

    According to `Becht et al.`_ the fold change is defined as

    .. math::
        FC = X - \overline{X}

    where :math:`X` is the mean of positive and :math:`\overline{X}` is the mean of the negative samples.

    Args:
        expr (np.ndarray): expression of `G` for each sample.
        positive_mask (np.ndarray, dtype=np.bool): boolean mask for `expr` indicating which samples belong to
            the positive class.

    Returns:
        float: fold change

    >>> expr = np.array([[2, 3, 5, 4, 9, 15]])
    >>> target = np.array(["A", "B", "C", "A", "B", "C"])
    >>> fold_change(expr, target == "A")
    array([-5.])

    .. _Becht et al.:
        http://dx.doi.org/10.1186/s13059-016-1070-5

    """
    return np.mean(expr[:, positive_mask], axis=1) - np.mean(
        expr[:, ~positive_mask], axis=1
    )


def specific_fold_change(expr, positive_mask, negative_masks):
    """
    Compute the specific fold change of the positive class with respect to all other classes
    for a single gene `G`.

    According to `Becht et al.`_ the specific fold change is defined as

    .. math::
        sFC = (X - \overline{X}_{min})/(\overline{X}_{max} - \overline{X}_{min})

    where :math:`X` is the mean of positive and :math:`\overline{X}_{max}` is the maximum mean over
    all negative classes and :math:`\overline{X}_{min}` is the minimal mean over all negative classes.

    Args:
        expr (np.ndarray): expression of `G` for each sample.
        positive_mask (np.ndarray): boolean mask for `expr` indicating which samples belong to
            the positive class.
        negative_masks (list of np.ndarray): list of boolean masks for `expr` indicating which samples belong
            to the different negative classes.

    Returns:
        float: specific fold change

    >>> expr = np.array([[2, 3, 5, 4, 9, 15], [2, 3, 5, 4, 9, 15]])
    >>> target = np.array(["A", "B", "C", "A", "B", "C"])
    >>> specific_fold_change(expr, target == 'A', [target == "B", target == "C"])
    array([-0.75, -0.75])

    .. _Becht et al.:
        http://dx.doi.org/10.1186/s13059-016-1070-5

    """
    mean_per_class = np.hstack(
        [
            np.mean(expr[:, class_inds], axis=1)[:, np.newaxis]
            for class_inds in negative_masks
        ]
    )
    x_min = np.min(mean_per_class, axis=1)
    x_max = np.max(mean_per_class, axis=1)
    return np.divide(np.mean(expr[:, positive_mask], axis=1) - x_min, x_max - x_min)


def roc_auc(expr, positive_mask):
    """
    Compute the Receiver Operator Characteristics Area under the Curve (ROC AUC) for a single gene `G`.
    This tells how well the gene discriminates between the two classes.

    This is a wrapper for the scikit-learn `roc_auc_score`_ function.

    Args:
       expr (np.ndarray): expression of `G` for each sample.
       positive_mask (np.ndarray, dtype=np.bool): boolean mask for `expr` indicating which samples belong to
            the positive class.

    Returns:
        float: roc auc score

    .. _roc_auc_score:
        http://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_auc_score.html#sklearn.metrics.roc_auc_score

    >>> expr = np.array([2, 3, 5, 4, 9, 15])
    >>> target = np.array(["A", "B", "C", "A", "B", "C"])
    >>> roc_auc(expr, target == "A")
    0.125

    """
    return roc_auc_score(positive_mask, expr)


class MCPSignatureGenerator(SignatureGenerator):
    """
    Implements the procedure described by `Becht et al.`_ for curating signatures.
    A gene is considered a valid `Transcription Marker` (TM) if it meets the following criteria

    * fold change >= ``min_fc`` (default=2)
    * specific fold change >= ``min_sfc`` (default=1.5)
    * AUC ROC >= ``min_auc`` (default=0.97)

    Args:
        expr: m x n gene expression matrix with m genes and n samples.
        target: m-vector with true tissue for each sample
        min_fc: minimal fold change for a gene to be considered as valid TM
        min_sfc: minimal specific fold change for a gene to be considered as valid TM
        min_auc: minimal ROC AUC for a gene to be considered as valid TM

    .. _Becht et al.:
        http://dx.doi.org/10.1186/s13059-016-1070-5

    """

    def __init__(self, expr, target, min_fc=2, min_sfc=1.5, min_auc=0.97):
        super().__init__(expr, target)
        self.min_fc = min_fc
        self.min_sfc = min_sfc
        self.min_auc = min_auc

    def _mk_signatures(self, expr, target):
        classes = list(set(target))
        masks = {cls: target == cls for cls in classes}
        signatures = {cls: [] for cls in classes}
        for cls in classes:
            fc = fold_change(expr, masks[cls])
            sfc = specific_fold_change(
                expr, masks[cls], [mask for k, mask in masks.items() if k != cls]
            )
            for i in range(expr.shape[0]):
                auc = roc_auc(expr[i, :], masks[cls])
                if (
                    fc[i] >= self.min_fc
                    and sfc[i] >= self.min_sfc
                    and auc >= self.min_auc
                ):
                    signatures[cls].append(i)

        return signatures


class MCPSignatureTester(SignatureTester):
    """
    Implements the `MCPCounter`_ described by Becht et al. in python.

    The principle is super-simple: take the mean of all marker genes as indicator.
    Also see their `R script`_.

    .. _R script:
        https://github.com/ebecht/MCPcounter/blob/a79614eee002c88c64725d69140c7653e7c379b4/Source/R/MCPcounter.R
    """

    def _score_signatures(self, expr, signatures):
        result = np.empty((len(signatures), expr.shape[1]))
        classes = self.sort_signatures(signatures)
        for i, cls in enumerate(classes):
            inds = np.array(list(signatures[cls]))
            for j in range(expr.shape[1]):
                result[i, j] = np.mean(expr[inds, j]) if len(inds) > 0 else np.NAN
        return result



from . import TESTDATA
import unittest
import pandas as pd
import pandas.util.testing as pdt
import numpy as np
from pygenesig.tools import translate_signatures
from pygenesig.gini import *


def sort_dict_of_lists(the_dict):
    """Make dict of unordered lists compareable."""
    return {key: sorted(value) for key, value in the_dict.items()}


class TestGini(unittest.TestCase):
    def test_gini_numeric_identity(self):
        """Check that the results are consistent with the rogini implementation up to four decimal digits."""
        df_aggr = pd.read_csv(
            TESTDATA / "./gini/input/expression_data.tsv", index_col=0, sep="\t"
        )
        rogini_res = pd.read_csv(
            TESTDATA / "./gini/output/expression_data_0.0_1.gini", sep="\t"
        )
        gini_res = df_aggr.apply(gini, 1)
        np.testing.assert_almost_equal(list(rogini_res.GINI_IDX), list(gini_res), 4)

    def test_get_rogini_format_00_1(self):
        """
        Check that the rogini output generated by pygenesig matches the 'original' by Laura.
        min_gini=0.0, max_rk=1
        """
        df_aggr = pd.read_csv(
            TESTDATA / "./gini/input/expression_data.tsv", index_col=0, sep="\t"
        )
        rogini_res = pd.read_csv(
            TESTDATA / "./gini/output/expression_data_0.0_1.gini", sep="\t"
        )
        rogini_res = rogini_res.iloc[:, :5]  # skip last column, as not compulsory.

        gini_format = get_rogini_format(df_aggr, min_gini=0, max_rk=1, min_expr=0)
        pdt.assert_frame_equal(
            rogini_res, gini_format, check_dtype=False, check_less_precise=True
        )

    def test_get_rogini_format_05_1(self):
        """
        Check that the rogini output generated by pygenesig matches the 'original' by Laura.
        min_gini=0.5, max_rk=1
        """
        df_aggr = pd.read_csv(
            TESTDATA / "./gini/input/expression_data.tsv", index_col=0, sep="\t"
        )
        rogini_res = pd.read_csv(
            TESTDATA / "./gini/output/expression_data_0.5_1.gini", sep="\t"
        )
        rogini_res = rogini_res.iloc[:, :5]  # skip last column, as not compulsory.

        gini_format = get_rogini_format(df_aggr, min_gini=0.5, max_rk=1, min_expr=0)
        pdt.assert_frame_equal(
            rogini_res, gini_format, check_dtype=False, check_less_precise=True
        )

    def test_get_rogini_format_06_3(self):
        """
        Check that the rogini output generated by pygenesig matches the 'original' by Laura.
        min_gini=0.6, max_rk=3
        """
        df_aggr = pd.read_csv(
            TESTDATA / "./gini/input/expression_data.tsv", index_col=0, sep="\t"
        )
        rogini_res = pd.read_csv(
            TESTDATA / "./gini/output/expression_data_0.6_3.gini", sep="\t"
        )
        rogini_res = rogini_res.iloc[:, :5]  # skip last column, as not compulsory.

        gini_format = get_rogini_format(df_aggr, min_gini=0.6, max_rk=3, min_expr=0)
        pdt.assert_frame_equal(
            rogini_res, gini_format, check_dtype=False, check_less_precise=True
        )

    def test_get_rogini_format_rank(self):
        """Check ranks in rogini format."""
        df_aggr = pd.read_csv(
            TESTDATA / "./gini/input/test_rank.tsv", index_col=0, sep="\t"
        )
        rogini_res = pd.read_csv(TESTDATA / "./gini/output/rank_0.5_3.gini", sep="\t")

        gini_format = get_rogini_format(df_aggr, min_gini=0.5, max_rk=3, min_expr=0)
        print(gini_format)
        pdt.assert_almost_equal(rogini_res, gini_format, check_dtype=False)

    def test_get_gini_signatures(self):
        """Check that the signatures derived with a simple groupby
        from rogini equals the signatures generated by pygenesig.gini."""
        df_aggr = pd.read_csv(
            TESTDATA / "./gini/input/expression_data.tsv", index_col=0, sep="\t"
        )
        rogini_res = pd.read_csv(
            TESTDATA / "./gini/output/expression_data_0.5_1.gini", sep="\t", index_col=0
        )
        grouped = rogini_res.groupby("CATEGORY")
        rogini_sig = {}
        for key, group in grouped:
            rogini_sig[key] = list(group.index)
        gini_sig = get_gini_signatures(df_aggr, min_gini=0.5, max_rk=1, min_expr=0)
        rosetta = dict(enumerate(df_aggr.index))
        gini_sig = translate_signatures(gini_sig, rosetta)
        self.assertDictEqual(
            sort_dict_of_lists(rogini_sig), sort_dict_of_lists(gini_sig)
        )

    def test_gini_gene_rank(self):
        """Check if row rank cutoff works."""
        df_aggr = pd.read_csv(
            TESTDATA / "./gini/input/test_rank.tsv", index_col=0, sep="\t"
        )
        rogini_res = pd.read_csv(TESTDATA / "./gini/output/rank_0.5_3.gini", sep="\t")

        signatures = get_gini_signatures(
            df_aggr, min_gini=0.5, max_rk=3, min_expr=0, max_rel_rk=0.25
        )
        expected = {"S1": set(), "S5": {3}, "S2": set(), "S3": set(), "S4": {3}}
        self.assertDictEqual(signatures, expected)



"""
With Differential Expression (DE)-Analysis one can find genes that
are significantly differentially expressed between different conditions.

This module provides a python bridge to the R packages `limma`_ and `edgeR`_
using ``rpy2`` and implements a ``SignatureGenerator`` based on DE.

.. _limma:
    https://bioconductor.org/packages/release/bioc/html/limma.html

.. _edgeR:
    https://www.bioconductor.org/packages/release/bioc/html/edgeR.html


"""

import readline
import rpy2.robjects as ro
from rpy2.robjects.packages import importr
from rpy2.robjects import numpy2ri, pandas2ri, Formula
from pygenesig.validation import SignatureTester
from pygenesig.file_formats import write_gmt
import sklearn.metrics
import numpy as np
from pygenesig.validation import SignatureGenerator
import tempfile

numpy2ri.activate()
pandas2ri.activate()

base = importr("base")
stats = importr("stats")
edge_r = importr("edgeR")
stats = importr("stats")
limma = importr("limma")


class LimmaSignatureGenerator(SignatureGenerator):
    """
    Use Differntial Expression (DE) Analysis to generate signatures.

    The LimmaSignatureGenerator uses `Limma` and `voom` to find significantly differentially
    expressed genes. Genes, which are specific for a tissue have a high fold-change and a low
    p-value, whereas genes equally present in all tissues will not be significant.

    The idea is, that genes which are significantly different between different tissues
    will reliably identify their tissue of origin.

    Args:
        expr (np.ndarray): m x n matrix with m samples and n genes
        target (array-like): m-vector with true tissue for each sample
        fold_change (int): minimal fold change for a gene to be considered as `differentially expressed`.
    """

    _TARGET_COL = "target"

    def __init__(self, expr, target, covariates, fold_change=100):
        super(LimmaSignatureGenerator, self).__init__(expr, target)
        assert (
            covariates.shape[0] == expr.shape[1]
        ), "nrow(covariates) must equal ncol(expr)"
        self.fold_change = fold_change
        self.covariates = covariates

    def _mk_signatures(self, expr, target):
        # genenames = indices
        genes = np.array([str(i) for i in range(expr.shape[0])])
        dg_list = edge_r.DGEList(counts=expr, genes=genes)

        assert all(
            [
                True if not col.startswith(self._TARGET_COL) else False
                for col in self.covariates.columns
            ]
        ), "covariates must not contain a column named target. "
        fmla = Formula("~ target + " + " + ".join(self.covariates.columns))
        env = fmla.environment
        env[self._TARGET_COL] = self.target
        for col in self.covariates.columns:
            env[col] = self.covariates[col]
        design_mat = stats.model_matrix(fmla)

        cpm = edge_r.cpm(dg_list)
        cpm_cnt = np.sum(np.array(cpm) > 1, axis=1)
        cpm_inds = cpm_cnt >= 10

        expr_fil = expr[cpm_inds, :]
        genes_fil = genes[cpm_inds]

        # overwrite with filtered dglist
        dg_list = edge_r.DGEList(counts=expr_fil, genes=genes_fil)
        dg_list = edge_r.calcNormFactors(dg_list)

        # run voom
        voom_res = limma.voom(dg_list, design_mat, plot=False)

        # run limma
        lm_fit = limma.lmFit(voom_res, design_mat)

        # filter based on pvalue and logfoldchange
        p_adj = 0.01 / base.nrow(lm_fit)[0]
        fit_e = limma.treat(lm_fit, lfc=np.log2(self.fold_change))
        rslt = limma.decideTests(fit_e, p_value=p_adj)

        # make signatures
        colnames = np.array(base.colnames(rslt))
        np_rslt = np.array(rslt)
        tissue_cols = np.array(
            [True if col.startswith("target") else False for col in colnames]
        )
        np_rslt = np_rslt[:, tissue_cols]
        tissues = np.array(
            [col[len(self._TARGET_COL) :] for col in colnames[tissue_cols]]
        )
        assert len(tissues) == np_rslt.shape[1]
        row_inds = np.where(cpm_inds)[0]
        signatures = {}
        for i, tissue in enumerate(tissues):
            signatures[tissue] = list(row_inds[np_rslt[:, i] == 1])

        return signatures



from pygenesig.validation import *
import unittest
import tempfile
import numpy as np
from dask import compute
import sklearn.model_selection


class TestValidation(unittest.TestCase):
    N_GENES = 1000
    N_SAMPELS = 200

    class DummySignatureGenerator(SignatureGenerator):
        DUMMY_SIGS = {"A": ("A", "B", "C"), "B": ("B", "C", "D")}

        def _mk_signatures(self, expr, target):
            return self.DUMMY_SIGS

    class DummySignatureTester(SignatureTester):
        def _score_signatures(self, expr, signatures):
            scores = np.zeros((len(signatures), expr.shape[1]))
            scores[0, :] = 1.0
            return scores

    def setUp(self):
        expr = np.random.random_sample((self.N_GENES, self.N_SAMPELS))
        target = np.array(
            ["A" if x < 0.3 else "B" for x in np.random.random_sample(self.N_SAMPELS)]
        )
        self.expr_file = tempfile.NamedTemporaryFile()
        self.target_file = tempfile.NamedTemporaryFile()
        np.save(self.expr_file, expr)
        np.savetxt(self.target_file, target, delimiter=",", fmt="%s")
        self.expr_file.flush()
        self.target_file.flush()

    def tearDown(self):
        self.expr_file.close()
        self.target_file.close()

    def test_cross_validation(self):
        sig_list, res_list, train_list, test_list = cv_score(
            self.expr_file.name,
            self.target_file.name,
            self.DummySignatureGenerator,
            self.DummySignatureTester,
            splitter=sklearn.model_selection.StratifiedKFold(n_splits=10),
        )
        # use the non-parallel scheduler for debugging.
        r_sig, r_res, r_train, r_test = compute(
            sig_list, res_list, train_list, test_list
        )
        self.assertEqual(10, len(r_sig))
        self.assertEqual(10, len(r_res))
        for i in range(10):
            self.assertEqual(self.DummySignatureGenerator.DUMMY_SIGS, r_sig[i])
            sm = r_res[i]
            self.assertEqual(
                sm.shape, (len(self.DummySignatureGenerator.DUMMY_SIGS), len(r_test[i]))
            )
            self.assertGreater(np.sum(sm), 0)

    def test_cross_validation_parallel(self):
        sig_list, res_list, train_list, test_list = cv_score(
            self.expr_file.name,
            self.target_file.name,
            self.DummySignatureGenerator,
            self.DummySignatureTester,
            splitter=sklearn.model_selection.StratifiedKFold(n_splits=10),
        )
        # use the threaded scheduler
        r_sig, r_res, r_train, r_test = compute(
            sig_list, res_list, train_list, test_list, scheduler="threads"
        )
        self.assertEqual(10, len(r_sig))
        self.assertEqual(10, len(r_res))
        for i in range(10):
            self.assertEqual(self.DummySignatureGenerator.DUMMY_SIGS, r_sig[i])
            sm = r_res[i]
            self.assertEqual(
                sm.shape, (len(self.DummySignatureGenerator.DUMMY_SIGS), len(r_test[i]))
            )
            self.assertGreater(np.sum(sm), 0)

    def test_filter_samples(self):
        target = np.array(["A"] * 2 + ["B"] * 10 + ["C"] * 4 + ["D"] * 14)
        exprs = np.random.rand(1000, 30)
        exprs2, target2 = filter_samples(exprs, target, n_splits=10)
        np.testing.assert_array_equal(target2, np.array(["B"] * 10 + ["D"] * 14))
        mask = np.array(list(range(2, 12)) + list(range(16, 30)))
        np.testing.assert_array_equal(exprs2, exprs[:, mask])



from pygenesig.validation import SignatureGenerator
from pygenesig.tools import collapse_matrix
import numpy as np
import pandas as pd
from collections import Counter
import itertools


class RankSignatureGenerator(SignatureGenerator):
    """
    Use gini index to generate gene signatures.

    Genes, which are specific for a tissue result in a high gini index,
    whereas genes equally present in all tissues have a gini index close to zero.

    The idea is, that genes with a high gini index will reliably identify
    their tissue of origin.

    Args:
        expr (np.ndarray): m x n matrix with m samples and n genes
        target (array-like): m-vector with true tissue for each sample
        min_gini (float): gini cutoff, genes need to have a gini index larger than this value.
        max_rk (int): rank cutoff, include genes if they rank ``<= max_rank`` among all tissues.
        min_expr (float): genes need to have at least an expression ``>= min_expr`` to be included.
        aggregate_fun (function): function used to aggregate samples of the same tissue.
    """

    def __init__(
        self,
        expr,
        target,
        include_fraction=0.1,
        not_in_fraction=0.5,
        max_signatures_per_gene=1,
    ):
        super(RankSignatureGenerator, self).__init__(expr, target)
        self.include_fraction = include_fraction
        self.not_in_fraction = not_in_fraction
        self.max_signatures_per_gene = max_signatures_per_gene

    def _mk_signatures(self, expr, target):
        df_aggr = collapse_matrix(expr, target, axis=1)
        include_index = np.array(range(int(self.include_fraction * df_aggr.shape[0])))
        not_in_index = np.array(range(int(self.not_in_fraction * df_aggr.shape[0])))
        tissue_series = {
            tissue: series.sort_values() for tissue, series in df_aggr.iteritems()
        }
        signatures = {
            tissue: series.iloc[include_index].index
            for tissue, series in tissue_series.items()
        }
        gene_count = Counter(
            itertools.chain.from_iterable(
                series.iloc[not_in_index].index for series in tissue_series.values()
            )
        )
        signatures_filtered = {
            tissue: [i for i in genes if gene_count[i] <= self.max_signatures_per_gene]
            for tissue, genes in signatures.items()
        }
        return signatures_filtered



"""
The Gini-coefficient (gini index) measures the inequality among values of
a frequency distribution. Originally applied to measure income inequality
we can also use it on gene expression data to find genes that are
over-represented in certain samples.
"""

import numpy as np
import pandas as pd
from pygenesig.validation import SignatureGenerator
from pygenesig.tools import collapse_matrix


def gini(array):
    """
    Calculate the Gini coefficient of a numpy array.

    Based on: https://github.com/oliviaguest/gini

    Args:
        array (array-like): input array

    Returns:
        float: gini-index of ``array``

    >>> a = np.zeros((10000))
    >>> a[0] = 1.0
    >>> '%.3f' % gini(a)
    '1.000'
    >>> a = np.ones(100)
    >>> '%.3f' % gini(a)
    '0.000'
    >>> a = np.random.uniform(-1,0,1000000)
    >>> '%.2f' % gini(a)
    '0.33'
    """
    # based on bottom eq: http://www.statsdirect.com/help/content/image/stat0206_wmf.gif
    # from: http://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm
    if np.amin(array) < 0:
        array -= np.amin(array)  # values cannot be negative
    array += 1e-12  # values cannot be 0
    array = np.sort(array)  # values must be sorted
    index = np.arange(1, array.shape[0] + 1)  # index per array element
    n = array.shape[0]  # number of array elements
    return (np.sum((2 * index - n - 1) * array)) / (
        n * np.sum(array)
    )  # Gini coefficient


def get_rogini_format(df_aggr, min_gini=0.7, max_rk=None, min_expr=1):
    """
    Imitate the *rogini* output format.

    Args:
        df_aggr (pd.DataFrame): m x n pandas DataFrame with m Genes and n tissues
        min_gini (float): gini cutoff, genes need to have a gini index larger than this value.
        max_rk (int): rank cutoff, include genes if they rank <= max_rank among all tissues.
        min_expr (float): minimal expression

    Returns:
        pd.DataFrame: equal to *rogini* output.

        Example::

            GENEID  CATEGORY        VALUE   RANKING GINI_IDX
            54165   Whole blood (ribopure)  491.359000      1       0.441296
            54165   CD34 cells differentiated to erythrocyte lineage        148.124000      2       0.441296
            54165   Mast cell - stimulated  68.973000       3       0.441296
            101927596       CD4+CD25-CD45RA+ naive conventional T cells     15.505000       1       0.948804
            101927596       CD8+ T Cells (pluriselect)      15.376000       2       0.948804
            101927596       CD4+CD25-CD45RA- memory conventional T cells    10.769000       3       0.948804

    """
    if max_rk is None:
        max_rk = float("inf")
    if min_gini is None:
        min_gini = 0
    if min_expr is None:
        min_expr = float("-inf")

    rogini_rows = []
    expr_gini = df_aggr.apply(gini, axis=1)
    expr_rank = df_aggr.rank(axis=1, ascending=False)
    for i in range(df_aggr.shape[0]):
        geneid = df_aggr.index[i]
        expr_row = df_aggr.iloc[i]
        gini_row = expr_gini.iloc[i]
        rk_row = expr_rank.iloc[i]
        which_cols = np.flatnonzero(
            (rk_row <= max_rk) & (gini_row >= min_gini) & (expr_row >= min_expr)
        )
        for j in which_cols:
            rogini_rows.append(
                [geneid, df_aggr.columns[j], expr_row[j], rk_row[j], gini_row]
            )
    columns = ["GENEID", "CATEGORY", "VALUE", "RANKING", "GINI_IDX"]
    df = pd.DataFrame(rogini_rows)
    df.columns = columns
    df.sort_values(["GENEID", "RANKING"], inplace=True)

    # the index does not have any meaning, we therefore want a consecutive index in the final data frame
    df.reset_index(drop=True, inplace=True)
    return df


def get_gini_signatures(
    df_aggr, min_gini=0.7, max_rk=3, max_rel_rk=None, min_expr=None
):
    """
    Generate gene signatures using gini index.

    Finds over-represented genes for each sample group, given the specified thresholds.

    Args:
        df_aggr (pd.DataFrame): m x n pandas DataFrame with m Genes and n tissues
        min_gini (float): gini cutoff, genes need to have a gini index larger than this value.
        max_rk (int): rank cutoff, include genes if they rank <= max_rank among all tissues.
        max_rel_rk (float): rank cutoff, include genes if they rank <= mal_rel_rk * m among all genes
            in the current sample.
        min_expr (float): minimal expression

    Returns:
        dict of list: A signature dictionary.

        Example::

            {
                "tissue1" : [list, of, gene, ids],
                "tissue2" : [list, of, other, genes],
                ...
            }

    """
    if max_rk is None:
        max_rk = float("inf")
    if min_gini is None:
        min_gini = 0
    if min_expr is None:
        min_expr = float("-inf")
    if max_rel_rk is None:
        max_rel_rk = 1

    # convert relative rank to absolute rank
    max_gene_rk = max_rel_rk * df_aggr.shape[0]

    expr_gini = df_aggr.apply(gini, axis=1)
    expr_rank_sample = df_aggr.rank(
        axis=1, ascending=False
    )  # rank across samples (along row)
    expr_rank_gene = df_aggr.rank(
        axis=0, ascending=False
    )  # rank across genes (along column)
    sig_mask = (
        (expr_rank_sample <= max_rk)
        & (df_aggr >= min_expr)
        & (expr_rank_gene <= max_gene_rk)
    )
    sig_mask.loc[
        expr_gini < min_gini,
    ] = False

    signatures = {}
    for tissue, sig_series in sig_mask.iteritems():
        signatures[tissue] = set(np.flatnonzero(sig_series))
    return signatures


class GiniSignatureGenerator(SignatureGenerator):
    """
    Use gini index to generate gene signatures.

    Genes, which are specific for a tissue result in a high gini index,
    whereas genes equally present in all tissues have a gini index close to zero.

    The idea is, that genes with a high gini index will reliably identify
    their tissue of origin.

    Args:
        expr (np.ndarray): m x n matrix with m samples and n genes
        target (array-like): m-vector with true tissue for each sample
        min_gini (float): gini cutoff, genes need to have a gini index larger than this value.
        max_rk (int): rank cutoff, include genes if they rank ``<= max_rank`` among all tissues.
        max_rel_rk (float): rank cutoff, include genes if they rank <= max_rel_rk * m among all genes
            in the current sample. Default: .33, i.e. genes need to appear in the first third of the sample.
        min_expr (float): genes need to have at least an expression ``>= min_expr`` to be included.
        aggregate_fun (function): function used to aggregate samples of the same tissue.
    """

    def __init__(
        self,
        expr,
        target,
        min_gini=0.7,
        max_rk=3,
        min_expr=1,
        max_rel_rk=0.33,
        aggregate_fun=np.median,
    ):
        super().__init__(expr, target)
        self.min_gini = min_gini
        self.max_rk = max_rk
        self.min_expr = min_expr
        self.aggregate_fun = aggregate_fun
        self.max_rel_rk = max_rel_rk

    def _mk_signatures(self, expr, target):
        df_aggr = collapse_matrix(
            expr, target, axis=1, aggregate_fun=self.aggregate_fun
        )
        return get_gini_signatures(
            df_aggr,
            min_gini=self.min_gini,
            max_rk=self.max_rk,
            min_expr=self.min_expr,
            max_rel_rk=self.max_rel_rk,
        )

    def get_rogini_format(self):
        df_aggr = collapse_matrix(
            self.expr, self.target, axis=1, aggregate_fun=self.aggregate_fun
        )
        return get_rogini_format(
            df_aggr, min_gini=self.min_gini, max_rk=self.max_rk, min_expr=self.min_expr
        )



import unittest
import numpy as np
import pandas as pd
import random
from pygenesig.tools import *
import pandas.util.testing as pdt


class TestTools(unittest.TestCase):
    def test_collapse_matrix_0(self):
        """test collapse matrix along axis 0"""
        mat = np.array(np.matrix("1 1 2 2;" "4 4 8 8;" "1 1 2 2;" "4 4 8 8;" "1 1 1 1"))
        group_by = ["A", "B", "A", "B", "C"]

        expected = pd.DataFrame(np.matrix("1 1 2 2;" "4 4 8 8;" "1 1 1 1"))
        expected.index = ["A", "B", "C"]

        actual = collapse_matrix(mat, group_by, axis=0)
        pdt.assert_frame_equal(actual, expected)

    def test_collapse_matrix_1(self):
        """test collapse matrix along axis 1"""
        mat = np.array(np.matrix("1 1 2 3;" "4 4 8 9;" "1 1 2 3;" "4 4 8 9;" "1 1 1 2"))
        group_by = ["A", "B", "A", "B"]

        expected = pd.DataFrame(np.matrix("3 4;" "12 13;" "3 4;" "12 13;" "2 3"))
        expected.columns = ["A", "B"]

        actual = collapse_matrix(mat, group_by, axis=1, aggregate_fun=np.sum)
        pdt.assert_frame_equal(actual, expected)

    def test_collapse_matrix_1_example2(self):
        """test collapse matrix along axis 1 with selecting a subset of columns."""
        expr = np.matrix("0 0 0 0;" "2 4 2 4;" "8 4 6 2")
        target = np.array(["A", "B", "A", "B"])
        expr_aggr = collapse_matrix(expr, target, axis=1)
        expr_aggr_expected = np.matrix("0 0;" "2 4;" "7 3")
        np.testing.assert_array_equal(expr_aggr_expected, expr_aggr)
        subset = np.array([0, 2, 3])
        expr_aggr_subset = collapse_matrix(expr[:, subset], target[subset], axis=1)
        expr_aggr_subset_expected = np.matrix("0 0;" "2 4;" "7 2")
        np.testing.assert_array_equal(expr_aggr_subset_expected, expr_aggr_subset)

    def test_collapse_matrix_chain(self):
        """re-aggregating an already aggregated matrix should return the identity."""
        nrow = 2000
        ncol = 1000
        expr = np.random.rand(nrow, ncol)
        target = np.random.choice(list("ABCDEFGHIJ"), ncol)
        mat_aggr = collapse_matrix(expr, target, axis=1)
        mat_aggr2 = collapse_matrix(mat_aggr.values, mat_aggr.columns, axis=1)
        pdt.assert_frame_equal(mat_aggr, mat_aggr2)



import numpy as np
import itertools
import pandas as pd
from pprint import pprint
from .tests import TESTDATA
from pygenesig.file_formats import load_gmt


def translate_signatures(signatures, rosetta, ignore_missing=False):
    """
    Translate gene identifiers in a signature dictionary.

    Args:
        signatures (dict of list): signature dictionary
        rosetta (dict): translation table mapping one gene identifier to another
        ignore_missing (boolean): If true, no error will be raised if an identifier is not in
            the translation dictionary. Respective entries will be skipped.

    Returns:
        dict of list: translated signature dictionary

    Raises:
        KeyError: if a gene is not in the rosetta dictionary unless ignore_missing is specified
    """
    if ignore_missing:
        # remove genes from signature which is not in rosetta.
        signatures = {
            tissue: [gene for gene in genes if gene in rosetta]
            for tissue, genes in signatures.items()
        }
    return {
        tissue: [rosetta[gene] for gene in genes]
        for tissue, genes in signatures.items()
    }


def combine_signatures(*args, function=set.intersection):
    """Combine signatures (e.g. by taking the intersection)
    Args:
        *args: list of signature dictonaries
        function: set operation to combine the signatures

    Returns:
        dict of set: combined signature dictionary.

    >>> s1 = {"A": {1, 2, 3}, "B": {2, 3, 4}}
    >>> s2 = {"A": {1, 3, 4, 5}, "B": {42}}
    >>> pprint(combine_signatures(s1, s2))
    {'A': {1, 3}, 'B': set()}
    """
    assert len(args) > 0, "No signatures provided"

    keys = args[0].keys()
    for sig in args:
        assert sig.keys() == keys, "All signature dictonaries must have identical keys."

    return {k: function(*[set(sig[k]) for sig in args]) for k in keys}


def jaccard_ind(set1, set2, *args):
    """
    Computes the Jaccard-Index of two or more sets.

    Args:
        set1 (list-like):
        set2 (list-like):
        *args: arbitrary number of more sets.

    Returns:
        float: jaccard index of all sets

    """
    set1 = set(set1)
    set2 = set(set2)
    i = len(set.intersection(set1, set2, *args))
    u = len(set.union(set1, set2, *args))
    return np.divide(i, u)


def pairwise_jaccard_ind(list_of_signatures):
    """
    Compute the pairwise jaccard index for a list
    of signature sets. Useful for calculating the pariwise overlap
    between the different crossvalidation folds.

    Args:
        list_of_signatures: list of signature dicts.

    Returns:
        dict: signature_name -> [list, of, jaccard, indices]

    Note:
        takes the signature names from the first dict in list_of_signatures
        to build the output dictionary.
    """
    assert len(list_of_signatures) > 0, "no signatures provided."
    pairwise_jacc = {}
    for signame in list_of_signatures[0]:
        pairwise_jacc[signame] = []
        for sigset1, sigset2 in itertools.product(
            list_of_signatures, list_of_signatures
        ):
            pairwise_jacc[signame].append(
                jaccard_ind(sigset1[signame], sigset2[signame])
            )
    return pairwise_jacc


def performance_per_tissue(list_of_confusion_matrices, sig_labels, perf_fun):
    """
    Compute per-tissue performance measures from all-against-all confusion matrices.

    Args:
        list_of_confusion_matrices (list of np.array):  list of confusion matrices
        sig_labels (array-like): list of signatures in the same order as in the confusion matrices.
        perf_fun (function): ``(TP, FN, TP, TN)`` computing a performance measure from the binary confusion matrix.
            See ``perfmeasures`` module.

    Returns:
        dict: signature_name -> list of performance meausures for each confusion matrix provided.

    """
    assert len(list_of_confusion_matrices) > 0, "no matrices provided."
    res = {}
    for i, sig in enumerate(sig_labels):
        res[sig] = []
        for confmat in list_of_confusion_matrices:
            TP = confmat[i, i]
            FN = np.sum(confmat[i, :]) - TP
            FP = np.sum(confmat[:, i]) - TP
            TN = np.sum(confmat) - TP - FN - FP
            res[sig].append(perf_fun(TP, FN, FP, TN))
    return res


def jaccard_mat(sigs1, sigs2, colname1="set_1", colname2="set_2", as_matrix=False):
    """
    Compute a matrix of jaccard indices to compute the overlap of two signature sets.

    Args:
        sigs1: signature dictionary
        sigs2: signature dictionary
        colname1: Name of the column for sigs1 in the dataframe
        colname2: Name of the column for sigs2 in the dataframe
        as_matrix: if False, a long-form dataframe will be returned, if True, a 2d matrix will be returned instead.

    Returns:
        pd.DataFrame: Matrix of Jaccard indices in long format

    Plot the overlap of signatures:
    >>> import seaborn as sns
    >>> signatures = load_gmt(TESTDATA / "bioqc/test_bioqc_log_pvalue.gmt")
    >>> df = jaccard_mat(signatures, signatures)
    >>> sns.heatmap(df.pivot(*df.columns))  # doctest: +ELLIPSIS
    <AxesSubplot:...>
    """
    jaccard_list = []
    for name1, genes1 in sigs1.items():
        for name2, genes2 in sigs2.items():
            jaccard_list.append((name1, name2, jaccard_ind(set(genes1), set(genes2))))

    df = pd.DataFrame(jaccard_list, columns=(colname1, colname2, "jaccard index"))
    if as_matrix:
        return df.pivot(index=colname1, columns=colname2)
    return df


def collapse_matrix(mat, group_by, axis=0, aggregate_fun=np.median):
    """
    Aggregate expression by annotation (collapse samples of the same tissue)

    Args:
        mat (np.array): m x n gene expression matrix with m genes and n samples.
        group_by (list-like): list of length m (if axis=0) or list of length n (if axis=1)
        axis (int): 0 for rows, 1 for columns
        aggregate_fun (function): aggregate to apply, defaults to ``numpy.median``

    Returns:
        pd.DataFrame: collapsed matrix with annotation from `group_by`.

    """
    if axis == 0:
        assert mat.shape[0] == len(group_by)
    elif axis == 1:
        assert mat.shape[1] == len(group_by)

    mat_df = pd.DataFrame(mat)
    group_by = list(group_by)  # strip index from series
    return mat_df.groupby(group_by, axis=axis).aggregate(aggregate_fun)


def normalize(array):
    """normalize a vector to values between 0 and 1"""
    amax = float(np.nanmax(array))
    amin = float(np.nanmin(array))
    if amax - amin == 0:
        return [0] * len(array)
    array = [(x - amin) / (amax - amin) for x in array]
    return array


def normalize_sum(array):
    """normalize a vector to sum to 1"""
    s = sum(array)
    return [x / s for x in array]



"""
`BioQC`_ is a R/Bioconductor package to detect tissue heterogeneity in gene expression data. It ships with a set
of 150 tissue signatures and implements a computationally efficient Wilcoxon-Mann-Whitney (WMW) test.

The WMW-test can be applied to check for an over-representation of an arbitrary gene set
in a gene expression sample.

This module implements a python bridge to the R package using ``rpy2`` and implements a
``SignatureTester`` based on BioQC.

.. _BioQC:
    https://accio.github.io/BioQC
"""


import readline  # not used, but needed as a workaround for an import problem.
import rpy2.robjects as ro
from rpy2.robjects.packages import importr
from rpy2.robjects import numpy2ri
from pygenesig.validation import SignatureTester
from pygenesig.file_formats import write_gmt
import sklearn.metrics
import numpy as np
import tempfile

numpy2ri.activate()

base = importr("base")
bioqc = importr("BioQC")
biobase = importr("Biobase")


wmw_test = bioqc.wmwTest
"""
Alias to R function ``BioQC::wmwTest`` using rpy2.

Check the `BioQC Documentation`_
for more details.

.. _BioQC Documentation:
    https://bioconductor.org/packages/release/bioc/manuals/BioQC/man/BioQC.pdf
"""

read_gmt = bioqc.readGmt
"""
Alias to R function ``BioQC::readGmt`` using rpy2.

Check the `BioQC Documentation`_
for more details.

.. _BioQC Documentation:
    https://bioconductor.org/packages/release/bioc/manuals/BioQC/man/BioQC.pdf
"""


ro.r(
    """
    mk_eset = function(exprs, gene_symbols) {
        return(new("ExpressionSet",
            exprs=exprs,
            featureData=new("AnnotatedDataFrame",
                            data.frame(GeneSymbol=gene_symbols))))
     }
     """
)
mk_eset = ro.r["mk_eset"]
"""
Generate an R ExpressionSet with ``exprs()`` and ``fData()`` annotation.
fData will contain one single column named GeneSymbol.

Args:
    exprs: ``numpy.array`` m x n expression matrix. ``np.matrix`` does NOT work!
    gene_symbols: ``numpy.array`` m vector containing gene symbols.
"""


class BioQCSignatureTester(SignatureTester):
    """
    Use BioQC to check the validity of gene signatures.

    We use the WMW-test implemented by BioQC to generate a p-value for
    each signature in each sample. If a signature is enriched in a specific sample
    the WMW-test will result in a small p-value.

    We define a sample as `correctly classified` by the signature
    if the signature corresponding to the `true tissue of origin`
    has the lowest p-value of all tested signatures.

    Args:
        expr (np.ndarray): m x n matrix with m samples and n genes
        target (array-like): m-vector with true tissue for each sample
    """

    @staticmethod
    def _reorder(template, target):
        """
        Reorder an array according to a template-array.

        Args:
            template: List of names how the order should be
            target: List of names how the order is

        Returns:
            np.ndarray: Ordered indices to bring target in the order or template.

        .. Note::
            ``template == target[_reorder(template, target)]`` is always ``True``.

        """
        target_array = np.array(target)
        return np.array([np.where(target_array == name)[0][0] for name in template])

    @staticmethod
    def signatures2gmt(signatures):
        """
        Convert signature dictionary into an R GMTList object.

        Args:
            signatures (dict): signature dictionary

        Returns:
            GMTList: R GMTList object containing the signatures.
        """
        with tempfile.NamedTemporaryFile() as gmt_file:
            write_gmt(signatures, gmt_file.name)
            gmt = read_gmt(gmt_file.name)
            return gmt

    @staticmethod
    def run_bioqc(expr, gene_symbols, gmt):
        """
        Apply BioQC to a gene expression matrix.

        Args:
            expr: m x n gene expression matrix with m genes and n samples.
            gene_symbols: m vector containing gene symbols
            gmt: GMTList object containing k signatures. Generate with ``signatures2gmt()``.

        Returns:
            np.array: k x n p-value matrix.
        """
        # although BioQC does support testing on a matrix and
        # list of signatures, I decided to go with the
        # gmt/eset solution. This is easier to debug and it's less easy to confuse things.
        eset = mk_eset(np.array(expr), np.array(gene_symbols))
        bioqc_res = wmw_test(eset, gmt, valType="p.greater")
        return np.array(bioqc_res)

    def _score_signatures(self, expr, signatures):
        gene_symbols = [str(x) for x in range(expr.shape[0])]
        signatures_not_empty = {
            name: genes for name, genes in signatures.items() if len(genes) > 0
        }
        gmt = BioQCSignatureTester.signatures2gmt(signatures_not_empty)
        bioqc_res = self.run_bioqc(expr, gene_symbols, gmt)
        if len(signatures_not_empty) == 1:
            # bug in R which returns a list instead of a matrix for one signature only.
            # make it a matrix again.
            bioqc_res = np.array([bioqc_res])
        bioqc_res_log = -np.log10(bioqc_res)

        reorder = self._reorder(list(base.names(gmt)), self.sort_signatures(signatures))
        # init empty result matrix
        result = np.empty((len(signatures), expr.shape[1]))
        result[:] = np.NAN
        # reassign row by row to result matrix to include empty signatures.
        for i_bioqc, i_result in enumerate(reorder):
            result[i_result, :] = bioqc_res_log[i_bioqc, :]

        return result



"""
This module contains high-level wrapper functions
for visualizing signature validation results.
"""

import pandas as pd
import numpy as np
from pylab import subplots
import seaborn as sns


def heatmap_figsize(nrow, ncol=None):
    """Generate size tuple for heatmap based on number of rows and columns"""
    if ncol is None:
        ncol = nrow
    return ncol * 0.3 + 3, nrow * 0.3


def aggregate_scores(sig_labels, scores, target):
    """
    Aggregate the scores of all samples of the same class.
    This is useful for plotting an average score heatmap.

    Args:
        sig_labels:
        scores:
        target:

    Returns:
        pd.DataFrame: n x m Data Frame with n signatures and m 'mean-of-target-class' columns.

    """
    scores_df = pd.DataFrame(np.transpose(scores))
    scores_df.columns = sig_labels
    scores_df = scores_df.assign(tissue=pd.Series(target))
    return scores_df.groupby("tissue").mean().transpose()


def plot_score_heatmap(heatmap_df, clip=30):
    """
    High-level function to plot a heatmap of a score matrix.

    Args:
        heatmap_df: Dataframe describing the heatmap
        clip: maximum color score (all values above will have the same color)

    Returns:
        fig: figure object
        ax: axis object

    """
    fig, ax = subplots(
        figsize=heatmap_figsize(heatmap_df.shape[0], heatmap_df.shape[1])
    )
    sns.heatmap(
        heatmap_df,
        ax=ax,
        annot=False,
        vmin=0,
        linewidths=0.2,
        vmax=clip,
        cbar_kws={"label": "BioQC score"},
    )
    ax.set_ylabel("signatures")
    ax.set_xlabel("mean of samples")
    return fig, ax


def plot_confusion_matrix_heatmap(confusion_matrix, labels):
    """
    Visualize a confusion matrix as an annotated heatmap.

    Args:
        confusion_matrix: n x n confusion matrix
        labels: n vector containing the axis labels

    Returns:
        fig: figure object
        ax: axis object

    """
    fig, ax = subplots(figsize=heatmap_figsize(confusion_matrix.shape[0]))
    sns.heatmap(
        confusion_matrix,
        ax=ax,
        xticklabels=labels,
        yticklabels=labels,
        annot=True,
        annot_kws={"size": 9},
        fmt=".0f",
        linewidths=0.2,
        vmin=0,
        vmax=100,
        cbar_kws={"label": "percent of class"},
    )
    ax.set_xlabel("predicted")
    ax.set_ylabel("actual")

    return fig, ax



import unittest
import numpy as np
import pandas as pd
import random
from pygenesig.tools import *
from pygenesig.file_formats import *
import pandas.util.testing as pdt
from tempfile import NamedTemporaryFile

np.random.seed(42)


def sort_dict_of_lists(the_dict):
    """Make dict of unordered lists compareable."""
    return {key: sorted(value) for key, value in the_dict.items()}


class TestTools(unittest.TestCase):
    def test_read_write_target(self):
        with NamedTemporaryFile(mode="w") as f:
            data = np.random.choice(["A", "B", "C", "D"], 100)
            write_target(data, f.name)
            read_data = read_target(f.name)
        np.testing.assert_array_equal(data, read_data)

    def test_read_write_rosetta(self):
        with NamedTemporaryFile(mode="w") as f:
            data = np.random.choice(["A", "B", "C", "D"], 100)
            write_rosetta(data, f.name)
            read_data = read_rosetta(f.name, as_dict=False)
        np.testing.assert_array_equal(data, read_data)

    def test_read_write_expr(self):
        with NamedTemporaryFile(mode="w", suffix=".npy") as f:
            data = np.random.rand(2000, 20)
            write_expr(data, f.name)
            read_data = read_expr(f.name)
        np.testing.assert_array_equal(data, read_data)

    def test_read_write_gct(self):
        with NamedTemporaryFile(mode="w") as f:
            data = np.random.rand(2000, 20)
            write_gct(f.name, data)
            read_data = read_gct(f.name)
        np.testing.assert_array_almost_equal(data, read_data)

    def test_read_write_gmt(self):
        with NamedTemporaryFile(mode="w") as f:
            signatures = {
                tissue: np.random.choice(["A", "B", "C", "D", "E", "F"], 3)
                for tissue in ["heart", "liver", "kidney", "brain"]
            }
            write_gmt(signatures, f.name)
            read_signatures = read_gmt(f.name)
        self.assertEqual(
            sort_dict_of_lists(signatures), sort_dict_of_lists(read_signatures)
        )



from distutils.core import setup
from setuptools import find_packages

# read the contents of your README file
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pygenesig",
    version="0.2.1",
    packages=find_packages(),
    description="Create and Validate Gene Signatures",
    author="Gregor Sturm",
    author_email="mail@gregor-sturm.de",
    url="https://github.com/grst/pygenesig",  # use the URL to the github repo
    keywords=["bioinformatics", "gene expression", "signatures"],
    license="MIT",
    install_requires=[
        "pyyaml",
        "seaborn",
        "numpy",
        "pandas",
        "rpy2",
        "scikit-learn",
        "dask",
        "distributed",
    ],
    extras_require=dict(
        tests=["nose", "black"],
        docs=["recommonmark", "sphinx", "sphinx_rtd_theme"],
    ),
    classifiers=[],
    include_package_data=True,
    long_description_content_type="text/x-rst",
    long_description=long_description,
)



import unittest
from pygenesig.mcp_counter import *
from pygenesig.validation import *
import logging


class TestMCPCounter(unittest.TestCase):
    def setUp(self):
        # np.matrix does NOT work l
        self.expr = np.array(
            np.matrix("5. 5. 0. 0.; " "0. 0. 5. 5.;" "3. 2. 2. 2.;" "0. 0. 0. 1.")
        )
        self.target = ["A", "A", "B", "B"]
        self.signatures = {  # signatures used in csv and gmt test files.
            "sig_a": [1, 2, 3, 4],
            "sig_b": [5, 6, 7, 8],
            "sig_c": [9, 10, 11, 12],
        }

    def test_roc_auc(self):
        expr = np.array([2, 3, 5, 4, 9, 15])
        target = np.array(["A", "B", "C", "A", "B", "C"])
        self.assertAlmostEquals(0.875, roc_auc(100 - expr, target == "A"))

    def test_score_signatures(self):
        signatures_bad = {"A": [3, 2], "B": [2]}
        st = MCPSignatureTester(self.expr, self.target)
        score = st.score_signatures(signatures_bad)
        score_expected = np.array(np.matrix("1.5 1.  1.  1.5;" "3.  2.  2.  2.  "))
        np.testing.assert_array_almost_equal(score_expected, score)

    def test_test_with_good_signatures(self):
        signatures_good = {"A": [0], "B": [1]}
        tester = MCPSignatureTester(self.expr, self.target)
        actual, predicted = tester._test_signatures(signatures_good, np.array(range(4)))
        np.testing.assert_array_equal(self.target, predicted)

    def test_test_with_bad_signatures_and_subset(self):
        signatures_bad = {"A": [3, 2], "B": [2]}
        tester = MCPSignatureTester(self.expr, self.target)
        actual, predicted = tester._test_signatures(signatures_bad, np.array([0, 1, 3]))
        cm = tester.confusion_matrix(signatures_bad, actual, predicted)
        cm_expected = np.matrix("0 2;" "0 1")
        np.testing.assert_array_equal(cm_expected, cm)

    def test_order_of_signatures(self):
        signatures_all = {"A": [0], "B": [1], "C": [3], "D": [3], "E": [3], "F": [3]}
        tester = MCPSignatureTester(self.expr, self.target)
        cm_expected = np.matrix(
            "2 0 0 0 0 0;"
            "0 2 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0"
        )
        for i in range(10):
            """test in loop to have random permutations of the dictionary."""
            new_dict = {k: v for k, v in signatures_all.items()}
            actual, predicted = tester._test_signatures(new_dict, np.array(range(4)))
            cm = tester.confusion_matrix(signatures_all, actual, predicted)
            np.testing.assert_array_equal(cm_expected, cm)

    def test_empty_signature_all(self):
        signatures_bad = {"A": [], "C": [], "B": []}
        tester = MCPSignatureTester(self.expr, self.target)
        with self.assertRaises(SignatureTesterException):
            actual, predicted = tester._test_signatures(
                signatures_bad, np.array([0, 1, 3])
            )

    def test_empty_signature1(self):
        signatures_bad = {"A": [3], "C": [], "B": []}
        tester = MCPSignatureTester(self.expr, self.target)
        actual, predicted = tester._test_signatures(signatures_bad, np.array([0, 1, 3]))
        cm = tester.confusion_matrix(signatures_bad, actual, predicted)
        cm_expected = np.matrix("2 0 0;" "1 0 0;" "0 0 0")
        np.testing.assert_array_equal(cm_expected, cm)

    def test_empty_signatures_order(self):
        signatures_all = {"A": [], "B": [2], "C": [3], "D": [], "E": [], "F": [3]}
        tester = MCPSignatureTester(self.expr, self.target)
        cm_expected = np.matrix(
            "0 2 0 0 0 0;"
            "0 2 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0;"
            "0 0 0 0 0 0"
        )
        for i in range(10):
            """test in loop to have random permutations of the dictionary."""
            new_dict = {k: v for k, v in signatures_all.items()}
            actual, predicted = tester._test_signatures(new_dict, np.array(range(4)))
            cm = tester.confusion_matrix(signatures_all, actual, predicted)
            np.testing.assert_array_equal(cm_expected, cm)

    def test_dtype(self):
        expr = np.array([["1.5", "2.7"], ["2.8", "3.7"]])
        target = np.array(["A", "B"])
        with self.assertRaises(TypeError):
            st = MCPSignatureTester(expr, target)



"""
Collection of functions to compute various performance measures from a 2x2 confusion matrix.
As an introduction to evaluating classificators, we recommend reading this `paper`_ about ROC analysis.

TP = true positive

FP = false positive

TN = true negative

FN = false negative

.. _paper: http://dx.doi.org/10.1016/j.patrec.2005.10.010.
"""


import numpy as np
from math import sqrt


def mcc(TP, FN, FP, TN):
    """Matthews Correlation Coefficient"""
    return np.divide(
        TP * TN + FP * FN, sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN))
    )


def sens(TP, FN, FP, TN):
    """Sensitivity"""
    return np.divide(TP, (TP + FN))


def spec(TP, FN, FP, TN):
    """Specificity"""
    return np.divide(TN, (TN + FP))


def prec_pos(TP, FN, FP, TN):
    """Posivitve Precision"""
    return np.divide(TP, (TP + FP))


def recall_pos(TP, FN, FP, TN):
    """Positive recall"""
    return np.divide(TP, (TP + FN))


def prec_neg(TP, FN, FP, TN):
    """Negative Precision"""
    return np.divide(TN, (TN + FN))


def recall_neg(TP, FN, FP, TN):
    """Negative recall"""
    return np.divide(TN, (TN + FP))


def f1_pos(TP, FN, FP, TN):
    """f1-measure on positive instances"""
    return np.divide(
        2 * prec_pos(TP, FN, FP, TN) * recall_pos(TP, FN, FP, TN),
        (prec_pos(TP, FN, FP, TN) + recall_pos(TP, FN, FP, TN)),
    )


def f1_neg(TP, FN, FP, TN):
    """f1-measure on negative instances."""
    return np.divide(
        2 * prec_neg(TP, FN, FP, TN) * recall_neg(TP, FN, FP, TN),
        (prec_neg(TP, FN, FP, TN) + recall_neg(TP, FN, FP, TN)),
    )
